#if !defined PLAYER_H
#define PLAYER_H

#include "config.h"

using namespace std;

/// <summary>
/// Struct used to hold the index of a 3-dimensional array.
/// </summary>
struct threeIndex {
	int x;
	int y;
	int z;
	threeIndex(int tx, int ty, int tz) : x(tx), y(ty), z(tz) {}
	threeIndex() : x(-1), y(-1), z(-1) {}
};

/// <summary>
/// Stores a Worker for the game Santorini.
/// <param name = "x"> x coordinate on the Board[x][j] </param>
/// <param name = "y"> y coordinate on the Board[x][j] </param>
/// <param name = "color"> Matches color of player it is associated with </param>
/// </summary>
struct Worker {
	//functions
	Worker() : x(-1), y(-1), color(0) {}
	Worker(int tx, int ty, int c) : x(tx), y(ty), color(c) {}

	bool getMoves(int m[8]);
	bool isMove();

	bool getPlacements(int m[8]);
	void makeMove(int move);
	void makeMove(int x,int y);
	void makePlacement(int move);
	
	//data members
	int x;
	int y;
	int color;
};

/// <summary>
/// Stores a player for the game Santorini.
/// <param name="name">Holds name of Player </param>
/// <param name="color"> Matches color of workers it is associated with </param>
/// </summary>
class Player {
public:
	//constructors
	Player() : name(""), color(0) {}
	Player(string nm, int col) : name(nm), color(col) {}

	int getColor() const;
	string getName() const;
	void setColor(int col);
	void setName(string nm);
	void getWorkers(int m[NUM_WORKERS_PER_PLAYER]);
	void getTransitions(bool m[NUM_WORKERS_PER_PLAYER][8][8]);
	void printTransitions();

	virtual int takeTurn() { return 0; };
	virtual void setWorker(int index) {};

protected:
	//data members
	string name;
	int color;
};

/// <summary>
/// Inherited class of Player for Human variety.
/// </summary>
class plyrHuman: public Player{
public:
	using Player::Player;
	virtual int takeTurn();
	virtual void setWorker(int index);
};

/// <summary>
/// Inherited class of Player for Random variety.
/// </summary>
class plyrRandom : public Player {
public:
	using Player::Player;
	virtual int takeTurn();
	virtual void setWorker(int index);
};

/// <summary>
/// Inherited class of Player for Simple AI ("Heuristic") variety.
/// </summary>
class plyrHeuristic : public Player {
public:
	using Player::Player;
	virtual int takeTurn();
	virtual void setWorker(int index);
	
	int printTurns();
	double EvalFcn(double v[50]);
	double getEval(double x[50]);

	int threshold = 4;
};

/// <summary>
/// Inherited class of Player for ANN.
/// </summary>
class plyrANN : public Player {
public:
	using Player::Player;
	plyrANN(string filename, string name = "", int color = 0);
	virtual int takeTurn();
	
	double EvalFcn(double v[50]);
	virtual void setWorker(int index);

	//weights
	double layer1[50][50];
	double layer2[50][50];
	double layer3[50][50];
	double layer4[1][50];

	//biases
	double shift1[50];
	double shift2[50];
	double shift3[50];

	double threshold = 0;

	void print(ostream& out = cout, string portion = "ALL");
	void ReLu(double v[50]);
	void shift(double v[50], int layer);
	void matrixMultiply(double v[50], int layer);
	double lastLayerMultiply(double v[50]);


	//training relevant functions.
	double getWeight(int layer, int i, int j);
	double getBias(int layer, int i);
	double getLogit(const double x[50], int layer, int i);
	double getHidden(const double x[50], int layer, int i);
	void updateANN(double g[7850]);
	void exportANN(string filename);
	double getEval(double x[50]);
};


//declares the existence of global arrays 
extern int Board[BOARD_SIZE][BOARD_SIZE];
extern Worker W[NUM_WORKERS];

int checkWorkers(int x, int y);
bool checkOccupied(int x, int y);
bool checkLevels(int x, int y, int x2, int y2);
bool checkFourth(int x, int y);
void exportBoardStateStream(int turn,ostream& out = cout);
void exportBoardState(int turn, string& out);
void importBoardState(int& turn, Player* p1, Player* p2, istream& in);
void importBoardState(int& turn, Player* p1, Player* p2, double v[50]);
void extractBoardState(int turn, double v[50]);
void printVector(double v[50]);

bool adjacentSpace(int move, int& x, int& y);
bool adjacentSpace(string move, int& x, int& y);
void printBoard(int space = 3, ostream& out = cout);
void setColor(int color);
void selectPos(int& x, int& y, string ask = "");

void validGameTree(double v[50], double tree[128]);
bool checkValidMove(int move, int index, double v[50]);
bool checkValidPlacement(int move, int index, double v[50]);
void buildGameTree(double v[50], double tree[128][50]);

double minimax(int depth, double board[50], bool isMaximizingPlayer, int maxDepth, plyrANN* p);
double minimax(int depth, double board[50], bool isMaximizingPlayer, int maxDepth, plyrHeuristic* p);

#endif